#include <Engine/types/Scene.h>
#include <Engine/types/entity/Camera.h>
#include <Engine/types/entity/InstancedSprite.h>
#include <JGL/JGL.h>
#include <algorithm>

using namespace Engine;
size_t Scene::EntityCount() const {
    return Entity::GetAllDescendents(entity_list).size();
}

void Scene::Update() {
    entity_list->UpdateChildren();
}

void Scene::Render(RenderTarget* render_target) {
    // The Camera and the HUD are ignored here because they're special cases.
    std::vector<Renderable*> display_list{};
    for (auto* e : Entity::GetAllDescendents(entity_list))
        if (auto* r = dynamic_cast<Renderable*>(e))
            if (r->ShouldRender())
                display_list.push_back(r);

    std::sort(display_list.begin(), display_list.end(), [](Renderable* a, Renderable* b) {
        return a->GetLayer() > b->GetLayer();
    });

    J2D::Begin( render_target, true);

    if (active_camera)
        active_camera->Render();

    for (auto& r : display_list)
        if (r->ShouldRender())
            r->Render();

    J2D::End();
}

Scene::~Scene() {
    for (auto* itx : instanced_textures)
        delete itx;

    for (auto* a : instanced_alpha_masks)
        delete a;

    delete entity_list;;
    delete active_camera;
}

void Scene::Unload() {
    for (auto* itx : instanced_textures)
        delete itx;

    for (auto* a : instanced_alpha_masks)
        delete a;

    delete entity_list;
    delete active_camera;
}

std::string Scene::GetName() const {
    return name;
}

Camera* Scene::GetActiveCamera() const {
    return active_camera;
}

InstancedTexture* Scene::GetInstancedTexture(const InstancedSprite* user) {
    for (auto* itx : instanced_textures)
        if (itx->InUseBy(user))
            return itx;

    auto* t = new Texture(user->GetTextureFilesystemPath());
    auto* itx = new InstancedTexture(t, user);
    instanced_textures.push_back(itx);
    return itx;
}

InstancedTexture* Scene::GetInstancedAlphaMask(const InstancedSprite* user) {
    for (auto* itx : instanced_alpha_masks)
        if (itx->InUseBy(user))
            return itx;

    if (!user->GetAlphaMaskFilesystemPath().empty()) {
        auto *t = new Texture(user->GetAlphaMaskFilesystemPath());
        auto *itx = new InstancedTexture(t, user);
        instanced_alpha_masks.push_back(itx);
        return itx;
    }
    return nullptr;
}

void Scene::DestroyInstancedTexture(const InstancedTexture* itx) {
    auto it = std::find(instanced_textures.begin(), instanced_textures.end(), itx);
    if (it != instanced_textures.end())
        delete *it, instanced_textures.erase(it);
}

void Scene::DestroyInstancedAlphaMask(const Engine::InstancedTexture* alpha_mask) {
    auto it = std::find(instanced_alpha_masks.begin(), instanced_alpha_masks.end(), alpha_mask);
    if (it != instanced_alpha_masks.end())
        delete *it, instanced_alpha_masks.erase(it);
}

